// ephemera/animator.cpp
// Animation controller: create and draw IFS fractals.
// Copyright (c) 2015 Ben Michell <http://www.muryan.com/>.

// This file is part of Ephemera.
//
// Ephemera is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Ephemera is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Ephemera.  If not, see <http://www.gnu.org/licenses/>.

#include "global.h"

#include "animator.h"
#include "random.h"


namespace eph
{


// Maximum number of times to try to create a fractal before giving up.
const int animator::m_max_fractal_attempts = 50;

// Number of times to try to create a fractal each call.
const int animator::m_per_call_fractal_attempts = 5;


//
// Construction.
//

animator::animator( unsigned char *bitmap, int width, int height, int bytes_per_row, int frame_rate ) :
	m_points_per_frame( std::max( 1000L, long( width ) * long( height ) / long( frame_rate ) / 4L ) ), 
	m_state( as_new_scheme ),
	m_renderer( bitmap, width, height, bytes_per_row ),
	m_current_scheme( 0 ),
	m_new_fractal_tries( 0 ), m_have_best_fractal( false ),
	m_target_fractals( 0 ), m_drawn_fractals( 0 ),
	m_target_points( 0 ),
	m_wait_time_remaining( 0 )
{
	// Define the schemes.
	scheme contractions( 10, 8, 12, true, cs_merge, 0.8, 0.7, 1.3, 0.7 );
	contractions.dimension_filter( 1.4 );
	contractions.min_saturation( 0.3 );
	contractions.min_brightness( 0.6 );
	contractions.add_rule_range( irt_linear, 2, 3 );
	m_schemes.push_back( contractions );

	scheme ethereal( 5, 1, 1, true, cs_intensity, 4.0, 0.8, 0.8, 0.0 );
	ethereal.dimension_filter( 1.5 );
	ethereal.min_saturation( 0.7 );
	ethereal.min_brightness( 0.9 );
	ethereal.add_rule_range( irt_linear, 2, 3 );
	m_schemes.push_back( ethereal );

	scheme shaded( 5, 3, 10, false, cs_merge, 0.6, 0.5, 1.6, 0.8 );
	shaded.min_saturation( 0.5 );
	shaded.min_brightness( 0.3 );
	shaded.add_rule_range( irt_linear, 2, 6 );
	shaded.add_rule_range( irt_trig, 2, 16 );
	shaded.add_rule_range( irt_trig_sq, 2, 16 );
	shaded.add_rule_range( irt_linear_trig, 2, 6 );
	shaded.add_rule_range( irt_signed_sqrt, 2, 12 );
	shaded.add_rule_range( irt_mixture, 3, 12 );
	m_schemes.push_back( shaded );

	scheme crystal( 5, 6, 12, false, cs_overwrite, 0.6, 0.4, 0.7, 0.9 );
	crystal.min_saturation( 0.6 );
	crystal.min_brightness( 0.9 );
	crystal.add_rule_range( irt_trig_sq, 6, 20 );
	crystal.add_rule_range( irt_signed_sqrt, 5, 16 );
	m_schemes.push_back( crystal );

	scheme mixture( 5, 5, 10, false, cs_merge, 0.6, 0.5, 1.6, 0.8 );
	mixture.add_rule_range( irt_mixture, 3, 12 );
	m_schemes.push_back( mixture );

	m_schemes.push_back( scheme( 3, st_ferns, 2 ) );
	m_schemes.push_back( scheme( 1, st_gasket, 2 ) );
}


//
// Advance the animation by one frame.
//

void
animator::update( long milliseconds )
{
	switch ( m_state )
	{
	default:
		assert( !"Unhandled animator state." );
		// Fall through.
	case as_new_scheme:
		choose_new_scheme();
		set_new_fractal_state();
		break;

	case as_new_fractal:
		if ( create_new_fractal() )
		{
			m_state = as_drawing_fractal;
		}
		break;

	case as_drawing_fractal:
		if ( draw_fractal() )
		{
			if ( ++m_drawn_fractals < m_target_fractals )
			{
				set_new_fractal_state();
			}
			else
			{
				m_state = as_post_draw_wait;
				m_wait_time_remaining = m_schemes[m_current_scheme].clear_delay() * 1000;
			}
		}
		break;

	case as_post_draw_wait:
		m_wait_time_remaining -= milliseconds;
		if ( m_wait_time_remaining <= 0 )
		{
			m_state = as_new_scheme;
		}
		break;
	}
}


//
// Select and prepare a new scheme for drawing.
//

void
animator::choose_new_scheme()
{
	m_renderer.clear();

	int total_weight = 0;
	for ( size_t s = 0; s < m_schemes.size(); ++s )
	{
		total_weight += m_schemes[s].weight();
	}
	int weight = random::get( 0, total_weight-1 );
	for ( m_current_scheme = 0; weight >= (int) m_schemes[m_current_scheme].weight(); ++m_current_scheme )
	{
		assert( m_current_scheme < m_schemes.size() );
		weight -= (int) m_schemes[m_current_scheme].weight();
	}
	assert( m_current_scheme < m_schemes.size() );

	m_target_fractals = random::get( m_schemes[m_current_scheme].min_fractals(), m_schemes[m_current_scheme].max_fractals() );
	m_drawn_fractals = 0;
}


//
// Switch to the new fractal state.
//

void
animator::set_new_fractal_state()
{
	m_state = as_new_fractal;
	m_new_fractal_tries = 0;
	m_have_best_fractal = false;
}


//
// Create a new fractal from the scheme.
//

bool
animator::create_new_fractal()
{
	ifs_fractal fractal;
	rectangle rect;
	double dimension;
	double area;
	bool found = false;
	int attempts = 0;
	do
	{
		// Look for a new fractal that meets the required dimension. While doing
		// this keep track of the best failed fractal.
		found = m_schemes[m_current_scheme].create_fractal( m_renderer.aspect_ratio(), fractal, rect, dimension, area );
		if ( !found && (!m_have_best_fractal || dimension > m_best_fractal_dimension) )
		{
			m_have_best_fractal = true;
			m_best_fractal = fractal;
			m_best_fractal_dimension = dimension;
			m_best_fractal_rect = rect;
			m_best_fractal_area = area;
		}

		// If we've tried unsuccessfully too many times then use the best fractal
		// found so far.
		if ( !found && (++m_new_fractal_tries >= m_max_fractal_attempts) )
		{
			assert( m_have_best_fractal );
			fractal = m_best_fractal;
			rect = m_best_fractal_rect;
			dimension = m_best_fractal_dimension;
			area = m_best_fractal_area;
			found = true;
		}
	}
	while ( !found && ++attempts < m_per_call_fractal_attempts );

	// If a suitable fractal was found install it into the renderer.
	if ( found )
	{
		m_renderer.fractal( fractal, rect, m_schemes[m_current_scheme].drawing_method() );
		m_target_points = long( m_schemes[m_current_scheme].point_density() * m_renderer.width() * m_renderer.height() * area );
	}
	return found;
}


//
// Draw some points for the current fractal.
//

bool
animator::draw_fractal()
{
	long draw = std::min( m_target_points, m_points_per_frame );
	m_renderer.draw( draw );
	m_target_points -= draw;
	return m_target_points == 0;
}


}; // namespace eph
