// ephemera/animator.h
// Animation controller: create and draw IFS fractals.
// Copyright (c) 2015 Ben Michell <http://www.muryan.com/>.

// This file is part of Ephemera.
//
// Ephemera is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Ephemera is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Ephemera.  If not, see <http://www.gnu.org/licenses/>.

#ifndef EPHEMERA_ANIMATOR_H
#define EPHEMERA_ANIMATOR_H


#include "ifs_fractal.h"
#include "renderer.h"
#include "scheme.h"


namespace eph
{


// Animator state machine.
enum animator_state
{
	as_new_scheme,
	as_new_fractal,
	as_drawing_fractal,
	as_post_draw_wait
};


//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


// Animation controller: create and draw IFS fractals.
class animator
{
public:

	animator( unsigned char *bitmap, int width, int height, int bytes_per_row, int frame_rate );
		// Construction.

	void update( long milliseconds );
		// Advance the animation by one frame. The milliseconds value gives the
		// time elapsed since the last call to this method.

private:

	void choose_new_scheme();
		// Select and prepare a new scheme for drawing.

	void set_new_fractal_state();
		// Switch to the new fractal state.

	bool create_new_fractal();
		// Create a new fractal from the scheme. The return value is true if a
		// new fractal has been found, otherwise this method should be called
		// again.

	bool draw_fractal();
		// Draw some points for the current fractal. The return value is true
		// when sufficient points have been drawn.

	const long m_points_per_frame;
		// Number of points to draw for each frame of animation. This helps to
		// keep the appearance of the screen saver relatively constant
		// regardless of the size of the screen it is running on.

	static const int m_max_fractal_attempts;
		// Maximum number of times to try to create a fractal before giving up.

	static const int m_per_call_fractal_attempts;
		// Number of times to try to create a fractal each call.

	animator_state m_state;
		// State machine.

	std::vector<scheme> m_schemes;
		// Collection of available schemes.

	renderer m_renderer;
		// Drawing management.

	size_t m_current_scheme;
		// Index of the currently selected scheme in the m_schemes collection.

	int m_new_fractal_tries;
	bool m_have_best_fractal;
	ifs_fractal m_best_fractal;
	double m_best_fractal_dimension;
	rectangle m_best_fractal_rect;
	double m_best_fractal_area;
		// Cached fractal during creation. When a dimension filter is active we
		// need to keep creating fractals until we find one that meets the
		// required dimension. But since this may be very hard (even bordering
		// on impossible) the number of attempts is limited, and when this limit
		// is reached we just use the best (highest dimension) one found.

	int m_target_fractals;
	int m_drawn_fractals;
		// Maximum and current fractals drawn.

	long m_target_points;
		// Number of points to draw for the current fractal.

	long m_wait_time_remaining;
		// Milliseconds remaining in the as_post_draw_wait state.
};


}; // namespace eph


#endif // EPHEMERA_ANIMATOR_H
