// ephemera/ephemera.cpp
// Ephemera screen saver.
// Copyright (c) 2015 Ben Michell <http://www.muryan.com/>.

// This file is part of Ephemera.
//
// Ephemera is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Ephemera is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Ephemera.  If not, see <http://www.gnu.org/licenses/>.

#include "global.h"

#include "animator.h"
#include "ephemera.h"
#include "ephemera.rh"
#include "version.h"


//
// Entry point.
//

int APIENTRY
WinMain( HINSTANCE, HINSTANCE, LPSTR cmd_line, int cmd_show )
{
	eph::ephemera saver;
	return saver.run_app( cmd_line, cmd_show, 30 );
}


namespace eph
{


//
// Construction.
//

ephemera::ephemera() :
	m_width( 0 ), m_height( 0 ),
	m_offscreen_dc( 0 ), m_offscreen_bitmap( 0 ), m_raw_bitmap( 0 ),
	m_animator( 0 ), m_last_animator_call( 0 )
{
}


//
// Destruction.
//

ephemera::~ephemera()
{
	if ( m_offscreen_dc != 0 )
	{
		DeleteDC( m_offscreen_dc );
		m_offscreen_dc = 0;
	}
	if ( m_offscreen_bitmap != 0 )
	{
		DeleteObject( m_offscreen_bitmap );
		m_offscreen_bitmap = 0;
		m_raw_bitmap = 0;
	}
	delete m_animator;
	m_animator = 0;
}


//
// Display the configuration dialog.
//

int
ephemera::configure( HWND parent )
{
	return DialogBox( 0, MAKEINTRESOURCE( IDD_CONFIGURE ), parent, ephemera::configure_dialog_proc );
}


//
// Initialize the screen saver.
//

bool
ephemera::saver_init( HWND hwnd, int width, int height )
{
	assert( m_animator == 0 );

	m_width = std::max( width, 1 );
	m_height = std::max( height, 1 );

	bool success = false;
	HDC hdc = GetDC( hwnd );
	if ( hdc != 0 )
	{
		m_offscreen_dc = CreateCompatibleDC( hdc );
		if ( m_offscreen_dc != 0 )
		{
			const BITMAPINFO bmi =
			{
				{ sizeof( BITMAPINFOHEADER ), m_width, -m_height, 1, 24, BI_RGB, 0, 0, 0, 0, 0 },
			};
			m_offscreen_bitmap = CreateDIBSection( hdc, &bmi, DIB_RGB_COLORS, (void **) &m_raw_bitmap, 0, 0 );
			if ( m_offscreen_bitmap != 0 )
			{
				SelectObject( m_offscreen_dc, m_offscreen_bitmap );

				RECT rect;
				SetRect( &rect, 0, 0, m_width, m_height );
				FillRect( m_offscreen_dc, &rect, (HBRUSH) GetStockObject( BLACK_BRUSH ) );
				GdiFlush();

				m_animator = new animator( m_raw_bitmap, m_width, m_height, (m_width*3 + 3) & ~3, frame_rate() );
				m_last_animator_call = GetTickCount();

				success = true;
			}
			else
			{
				DeleteDC( m_offscreen_dc );
				m_offscreen_dc = 0;
			}
		}
		ReleaseDC( hwnd, hdc );
	}
	return success;
}


//
// Draw the window while running as a screen saver.
//

void
ephemera::saver_draw( HDC hdc, const RECT &dirty )
{
	if ( m_offscreen_dc != 0 && m_offscreen_bitmap != 0 )
	{
		BitBlt( hdc, 0, 0, dirty.right, dirty.bottom, m_offscreen_dc, 0, 0, SRCCOPY );
	}
	else
	{
		FillRect( hdc, &dirty, (HBRUSH) GetStockObject( BLACK_BRUSH ) );
	}
}


//
// Perform animation processing.
//

bool
ephemera::saver_animate()
{
	if ( m_animator != 0 )
	{
		DWORD now = GetTickCount();
		long elapsed = std::max( long( now - m_last_animator_call ), 0L );
		m_last_animator_call = now;
		m_animator->update( elapsed );
		return true;
	}
	return false;
}


//
// Dialog proc for the configuration dialog.
//

BOOL CALLBACK
ephemera::configure_dialog_proc( HWND hwnd, UINT msg, WPARAM wparam,  LPARAM lparam )
{
	switch ( msg )
	{
	case WM_INITDIALOG:
		centre_over_parent( hwnd );
		SetDlgItemText( hwnd, IDC_VERSION, EPH_VERSION_STRING );
		return TRUE; 

	case WM_COMMAND:
		switch ( LOWORD( wparam ) )
		{
		case IDOK:
		case IDCANCEL:
			if ( HIWORD( wparam ) == BN_CLICKED )
			{
				EndDialog( hwnd, wparam );
				return TRUE;
			}
			break;

		default:
			break;
		}
	}
	return FALSE;
}


}; // namespace eph
