// ephemera/ifs_fractal.cpp
// IFS fractal.
// Copyright (c) 2015 Ben Michell <http://www.muryan.com/>.

// This file is part of Ephemera.
//
// Ephemera is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Ephemera is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Ephemera.  If not, see <http://www.gnu.org/licenses/>.

#include "global.h"

#include "ifs_fractal.h"
#include "random.h"


namespace eph
{


// Number of points to generate when analyzing the fractal.
const size_t ifs_fractal::m_analysis_points = 10000;

// Granularity of the coarse grid when computing the box-counting
// dimension.
const size_t ifs_fractal::m_box_counting_granularity = 32;


//
// Create a randomized set of rules.
//

void
ifs_fractal::randomize( ifs_rule_type rule_type, int min_rules, int max_rules, bool contractive, const double &min_saturation, const double &min_brightness )
{
	assert( rule_type >= 0 );
	assert( rule_type < irt_count );
	assert( min_rules >= 2 );
	assert( max_rules >= min_rules );
	assert( min_saturation >= 0.0 );
	assert( min_saturation <= 1.0 );
	assert( min_brightness >= 0.0 );
	assert( min_brightness <= 1.0 );

	size_t rules = (size_t) random::get( min_rules, max_rules );
	m_rules.clear();
	while ( m_rules.size() < rules )
	{
		ifs_rule_type new_type = rule_type;
		if ( new_type == irt_mixture )
		{
			new_type = (ifs_rule_type) random::get( 0, irt_normal_count-1 );
		}

		rgb colour;
		colour.from_hsv( random::get( 0.0, 1.0 ), random::get( min_saturation, 1.0 ), random::get( min_brightness, 1.0 ) );
		ifs_rule rule( 1, new_type, 0, 0, 0, 0, 0, 0, colour, 0.333 );
		const int max_tests = 100;
		int test = 0;
		do
		{
			rule.a( random::get( -1.0, 1.0 ) );
			rule.b( random::get( -1.0, 1.0 ) );
			rule.c( random::get( -1.0, 1.0 ) );
			rule.d( random::get( -1.0, 1.0 ) );
			rule.e( random::get( -1.0, 1.0 ) );
			rule.f( random::get( -1.0, 1.0 ) );
		}
		while ( ++test < max_tests && contractive && rule.contractivity() == irc_not_contractive );
		add_rule( rule );
	}
}


//
// Scale the fractal by the specified amounts.
//

void
ifs_fractal::scale( const double &scale_x, const double &scale_y )
{
	for ( std::vector<ifs_rule>::iterator iter = m_rules.begin(); iter != m_rules.end(); ++iter )
	{
		iter->scale( scale_x, scale_y );
	}
}


//
// Shear the fractal by the specified amounts.
//

void
ifs_fractal::shear( const double &x_in_y, const double &y_in_x )
{
	for ( std::vector<ifs_rule>::iterator iter = m_rules.begin(); iter != m_rules.end(); ++iter )
	{
		iter->shear( x_in_y, y_in_x );
	}
}


//
// Rotate the fractal by the specified angle.
//

void
ifs_fractal::rotate( const double &radians )
{
	for ( std::vector<ifs_rule>::iterator iter = m_rules.begin(); iter != m_rules.end(); ++iter )
	{
		iter->rotate( radians );
	}
}


//
// Mirror the fractal in X.
//

void
ifs_fractal::mirror_x()
{
	for ( std::vector<ifs_rule>::iterator iter = m_rules.begin(); iter != m_rules.end(); ++iter )
	{
		iter->mirror_x();
	}
}


//
// Mirror the fractal in Y.
//

void
ifs_fractal::mirror_y()
{
	for ( std::vector<ifs_rule>::iterator iter = m_rules.begin(); iter != m_rules.end(); ++iter )
	{
		iter->mirror_y();
	}
}


//
// Adjust the size and position of the fractal.
//

void
ifs_fractal::scale_and_move( rectangle &rect, double min_scale, double max_scale, double centre )
{
	assert( rect.width() > 0.0 );
	assert( rect.height() > 0.0 );
	assert( epsilon_pos( min_scale ) );
	assert( max_scale >= min_scale );
	assert( centre >= 0.0 );

	// First scale. Selecting a scale with logs evens out the selected values in
	// an aesthetically pleasing way.
	double min_scale_log = log( min_scale );
	double max_scale_log = log( max_scale );
	double scale = exp( random::get( log( min_scale ), log( max_scale ) ) );

	double radius_x = rect.width() * 0.5;
	double centre_x = rect.min_x() + radius_x;
	radius_x /= scale;
	rect.min_x( centre_x - radius_x );
	rect.max_x( centre_x + radius_x );

	double radius_y = rect.height() * 0.5;
	double centre_y = rect.min_y() + radius_y;
	radius_y /= scale;
	rect.min_y( centre_y - radius_y );
	rect.max_y( centre_y + radius_y );

	// Move the centre.
	double offset_x = radius_x * random::get( -centre, centre );
	double offset_y = radius_y * random::get( -centre, centre );
	rect.offset( point( offset_x, offset_y ) );
}


//
// Analyze the IFS and return its range and fractal dimension.
//

void
ifs_fractal::analyze( rectangle &rect, double &dimension ) const
{
	// Obtain a start point for point generation, and use that as the initial
	// range.
	assert( !m_rules.empty() );
	ifs_point pt;
	start_point( pt );
	rect.min( pt.pt() );
	rect.max( pt.pt() );

	// Ensure any known fixed points are covered.
	for ( size_t rule = 1; rule < m_rules.size(); ++rule )
	{
		ifs_point fixed_pt;
		if ( m_rules[rule].fixed_point( fixed_pt ) )
		{
			rect.include( fixed_pt.pt() );
		}
	}

	// Run off points to give a good idea of the fractal's range.
	std::vector<point> pts( m_analysis_points );
	pts[0] = pt.pt();
	size_t p;
	for ( p = 1; p < m_analysis_points; ++p )
	{
		m_rules[random::get( 0U, m_rules.size()-1 )].apply( pts[p-1], pts[p] );
		rect.include( pts[p] );
	}

	// Calculate the box-counting dimension. First ensure the range is adequate.
	double width = rect.width();
	double height = rect.height();
	dimension = 0.0;
	if ( epsilon_pos( width ) && epsilon_pos( height ) )
	{
		// Two sets of boxes are used: one set with the given granularity and
		// another with twice as many.
		const size_t coarse_granularity = m_box_counting_granularity;
		const size_t fine_granularity = coarse_granularity * 2;

		// Compute the scaling from the fractal to the fine boxes.
		double scale_x = fine_granularity / width;
		double scale_y = fine_granularity / height;
			
		// Allocate bit sets for a grid of boxes over the fractal. This is done
		// with raw bit flags instead of vector<bool> since the latter is
		// horrendously slow.
		std::vector<unsigned char> coarse( (coarse_granularity * coarse_granularity + 7U) >> 3U, 0 );
		std::vector<unsigned char> fine( (fine_granularity * fine_granularity + 7U) >> 3U, 0 );

		// Count boxes using the buffered points.
		unsigned long coarse_count = 0;
		unsigned long fine_count = 0;
		unsigned long bx;
		unsigned long by;
		size_t flag_offset;
		for ( p = 0; p < pts.size(); ++p )
		{
			// Convert the point into a fine box array index. Note that a
			// negative (out of range) value will become an out of range
			// positive value.
			bx = (unsigned long) (long) floor( (pts[p].x() - rect.min_x()) * scale_x );
			by = (unsigned long) (long) floor( (pts[p].y() - rect.min_y()) * scale_y );
				
			// Use the point if it is in range.
			if ( bx < fine_granularity && by < fine_granularity )
			{
				// Add a point to the fine box grid if not already set.
				flag_offset = bx + by * fine_granularity;
				size_t flag_byte = flag_offset >> 3U;
				unsigned char flag_bit = (unsigned char) (1U << (flag_offset & 0x7));
				if ( (fine[flag_byte] & flag_bit) == 0 )
				{
					fine[flag_byte] |= flag_bit;
					fine_count++;
				}
					
				// Add a point to the coarse box grid if not already set.
				bx >>= 1;
				by >>= 1;
				flag_offset = bx + by * coarse_granularity;
				flag_byte = flag_offset >> 3U;
				flag_bit = (unsigned char) (1U << (flag_offset & 0x7));
				if ( (coarse[flag_byte] & flag_bit) == 0 )
				{
					coarse[flag_byte] |= flag_bit;
					coarse_count++;
				}
			}
		}

		// At least one hit is required to be able to compute the dimension.
		if ( fine_count > 0 )
		{
			assert( fine_count >= coarse_count );
			dimension = log( (double) fine_count / (double) coarse_count ) / log( 2.0 );
		}
	}
}


//
// Find a starting point to use for this fractal.
//

void
ifs_fractal::start_point( ifs_point &pt ) const
{
	assert( !m_rules.empty() );
	for ( size_t rule = 0; rule < m_rules.size(); ++rule )
	{
		if ( m_rules[rule].fixed_point( pt ) )
		{
			return;
		}
	}
	if ( !m_rules.empty() )
	{
		m_rules[0].start_point( pt );
	}
	else
	{
		pt.pt().xy( 0, 0 );
		pt.colour( rgb( 0, 0, 0 ) );
	}
}


//
// Generate a set of points.
//

void
ifs_fractal::generate_points( const ifs_point &start_point, std::vector<ifs_point> &points, size_t point_count ) const
{
	assert( !m_rules.empty() );
	points.resize( point_count );
	if ( point_count > 0 )
	{
		// Total the weights of the rules.
		int total_weight = 0;
		bool equal_weights = true;
		size_t r;
		for ( r = 0; r < m_rules.size(); ++r )
		{
			total_weight += m_rules[r].weight();
			if ( r > 0 && equal_weights )
			{
				equal_weights = m_rules[r].weight() == m_rules[r-1].weight();
			}
		}

		// If all weights are equal then select rules with a simple choice.
		if ( equal_weights )
		{
			m_rules[random::get( 0U, m_rules.size()-1 )].apply( start_point, points[0] );
			for ( size_t p = 1; p < point_count; ++p )
			{
				m_rules[random::get( 0U, m_rules.size()-1 )].apply( points[p-1], points[p] );
			}
		}
		// Otherwise use a weighted choice.
		else
		{
			int w = random::get( 0, total_weight-1 );
			for ( r = 0; w >= m_rules[r].weight(); ++r )
			{
				assert( r < m_rules.size() );
				w -= m_rules[r].weight();
			}
			assert( r < m_rules.size() );
			m_rules[r].apply( start_point, points[0] );
			for ( size_t p = 1; p < point_count; ++p )
			{
				w = random::get( 0, total_weight-1 );
				for ( r = 0; w >= m_rules[r].weight(); ++r )
				{
					assert( r < m_rules.size() );
					w -= m_rules[r].weight();
				}
				assert( r < m_rules.size() );
				m_rules[r].apply( points[p-1], points[p] );
			}
		}
	}
}


}; // namespace eph
