// ephemera/ifs_fractal.h
// IFS fractal.
// Copyright (c) 2015 Ben Michell <http://www.muryan.com/>.

// This file is part of Ephemera.
//
// Ephemera is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Ephemera is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Ephemera.  If not, see <http://www.gnu.org/licenses/>.

#ifndef EPHEMERA_IFS_FRACTAL_H
#define EPHEMERA_IFS_FRACTAL_H


#include "ifs_rule.h"


namespace eph
{


// IFS fractal.
class ifs_fractal
{
public:

	ifs_fractal() {}
		// Construction.

	void clear() { m_rules.clear(); }
		// Remove all rules from the fractal.

	bool empty() const { return m_rules.empty(); }
		// Test if the fractal is empty (has no rules).

	void add_rule( const ifs_rule &rule ) { m_rules.push_back( rule ); }
		// Add a rule to the fractal.

	void randomize( ifs_rule_type rule_type, int min_rules, int max_rules, bool contractive, const double &min_saturation, const double &min_brightness );
		// Create a randomized set of rules. The rule type can be one of the
		// normal values or irt_mixture. This latter value will generate a mix
		// of all other rule types. The contractive flag only works with the
		// linear rule type.

	void scale( const double &scale_x, const double &scale_y );
		// Scale the fractal by the specified amounts. This only works with the
		// linear rule type.

	void shear( const double &x_in_y, const double &y_in_x );
		// Shear the fractal by the specified amounts. This only works with the
		// linear rule type.

	void rotate( const double &radians );
		// Rotate the fractal by the specified angle. This only works with the
		// linear rule type.

	void mirror_x();
	void mirror_y();
		// Mirror the fractal in X or Y. This only works with the linear rule
		// type.

	void scale_and_move( rectangle &rect, double min_scale, double max_scale, double centre );
		// Adjust the size and position of the fractal. A scale value will be
		// chosen evenly within the given limits. The position of the fractal
		// within the rectangle will then be adjusted by up to the centre
		// amount, where a value of 0 means not at all and a value of 1 means
		// the edge of the rectangle.

	void analyze( rectangle &rect, double &dimension ) const;
		// Analyze the IFS and return its range and fractal dimension.

	void start_point( ifs_point &pt ) const;
		// Find a starting point to use for this fractal. If any of the rules
		// has a fixed point then this is used, otherwise a point on the
		// attractor will be determined empirically.

	void generate_points( const ifs_point &start_point, std::vector<ifs_point> &points, size_t point_count ) const;
		// Generate a set of points.

private:

	static const size_t m_analysis_points;
		// Number of points to generate when analyzing the fractal.

	static const size_t m_box_counting_granularity;
		// Granularity of the coarse grid when computing the box-counting
		// dimension.

	std::vector<ifs_rule> m_rules;
		// Rules making up the fractal.
};


}; // namespace eph


#endif // EPHEMERA_IFS_FRACTAL_H
