// ephemera/ifs_rule.cpp
// A rule in an IFS fractal.
// Copyright (c) 2015 Ben Michell <http://www.muryan.com/>.

// This file is part of Ephemera.
//
// Ephemera is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Ephemera is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Ephemera.  If not, see <http://www.gnu.org/licenses/>.

#include "global.h"

#include "ifs_rule.h"
#include "matrix3.h"


namespace eph
{


//
// Construction.
//

ifs_rule::ifs_rule( int weight, ifs_rule_type rule_type, const double &a, const double &b, const double &c, const double &d, const double &e, const double &f, const rgb &colour, const double &blend ) :
	m_weight( weight ),
	m_rule_type( rule_type ),
	m_a( a ), m_b( b ), m_c( c ), m_d( d ), m_e( e ), m_f( f ),
	m_colour( colour )
{
	assert( weight > 0 );

	// Precompute the colour tables.
	const double dr = double( colour.r() )*blend;
	const double dg = double( colour.g() )*blend;
	const double db = double( colour.b() )*blend;
	for ( int c = 0; c < 256; ++c )
	{
		const double dc = double( c ) * (1.0 - blend);
		m_red[c] = (unsigned char) clamp( int( dr + dc + 0.5 ), 0, 255 );
		m_green[c] = (unsigned char) clamp( int( dg + dc + 0.5 ), 0, 255 );
		m_blue[c] = (unsigned char) clamp( int( db + dc + 0.5 ), 0, 255 );
	}
}


//
// Scale the rule by the specified amounts.
//

void
ifs_rule::scale( const double &scale_x, const double &scale_y )
{
	assert( m_rule_type == irt_linear );
	assert( epsilon_nz( scale_x ) );
	assert( epsilon_nz( scale_y ) );

	// Scaling works by combining the transformations to unscale image space
	// into rule space, applying the original rule, then scaling back into image
	// space:
	//
	// [ x' y' 1 ] = [ x y 1 ] . [ scale 1/s ] . [ rule ] . [ scale s ]
	if ( epsilon_nz( scale_x ) && epsilon_nz( scale_y ) )
	{
		matrix3	m;
		m.scale( 1.0/scale_x, 1.0/scale_y );
		matrix3	temp;
		to_matrix( temp );
		m = m * temp;
		temp.scale( scale_x, scale_y );
		m = m * temp;
		from_matrix( m );
	}
}


//
// Rotate the rule by the specified angle.
//

void
ifs_rule::rotate( const double &radians )
{
	assert( m_rule_type == irt_linear );

	// Rotation works by combining the transformations to unrotate image space
	// into rule space, applying the original rule, then rotating back into
	// image space:
	//
	// [ x' y' 1 ] = [ x y 1 ] . [ rotate -r ] . [ rule ] . [ rotate r ]
	matrix3	m;
	m.rotate( -radians );
	matrix3	temp;
	to_matrix( temp );
	m = m * temp;
	temp.rotate( radians );
	m = m * temp;
	from_matrix( m );
}


//
// Shear the rule by the specified amounts.
//

void
ifs_rule::shear( const double &x_in_y, const double &y_in_x )
{
	assert( m_rule_type == irt_linear );

	// Shearing works by combining the transformations to unshear image space
	// into rule space, applying the original rule, then shearing back into
	// image space:
	//
	// [ x' y' 1 ] = [ x y 1 ] . [ shear -s ] . [ rule ] . [ shear s ]
	matrix3	m;
	m.shear( -x_in_y, -y_in_x);
	matrix3	temp;
	to_matrix( temp );
	m = m * temp;
	temp.shear( x_in_y, y_in_x );
	m = m * temp;
	from_matrix( m );
}


//
// Apply the rule to the coloured point.
//

void
ifs_rule::apply( const ifs_point &pt, ifs_point &new_point ) const
{
	// Transform the point.
	apply( pt.pt(), new_point.pt() );

	// Transform the colour.
	new_point.colour().r( m_red[pt.colour().r()] );
	new_point.colour().g( m_green[pt.colour().g()] );
	new_point.colour().b( m_blue[pt.colour().b()] );
}


//
// Apply the rule to the point.
//

void
ifs_rule::apply( const point &pt, point &new_pt ) const
{
	// Transform the point.
	switch ( m_rule_type )
	{
	default:
		assert( !"Invalid rule system." );
		// Fall through.
	case irt_linear:
		new_pt.xy( pt.x()*m_a + pt.y()*m_b + m_e, pt.x()*m_c + pt.y()*m_d + m_f );
		break;

	case irt_trig:
		new_pt.xy( cos( pt.x() )*m_a + cos( pt.y() )*m_b + m_e, sin( pt.x() )*m_c + sin( pt.y() )*m_d + m_f );
		break;

	case irt_trig_sq:
		{
			double cx = cos( pt.x() );
			double sx = sin( pt.x() );
			double cy = cos( pt.y() );
			double sy = sin( pt.y() );
			new_pt.xy( cx*cx*m_a + cy*cy*m_b + m_e, sx*sx*m_c + sy*sy*m_d + m_f );
		}
		break;

	case irt_linear_trig:
		{
			double cx = cos( pt.x() );
			double sy = sin( pt.y() );
			new_pt.xy( pt.x()*m_a + pt.y()*m_b + m_e, cx*cx*m_c + sy*sy*m_d + m_f );
		}
		break;

	case irt_signed_sqrt:
		{
			double ssqrt_x = ssqrt( pt.x() );
			double ssqrt_y = ssqrt( pt.y() );
			new_pt.xy( ssqrt_x*m_a + ssqrt_y*m_b + m_e, ssqrt_x*m_c + ssqrt_y*m_d + m_f );
		}
		break;
	}
}


//
// Find a starting point to use for this rule.
//

void
ifs_rule::start_point( ifs_point &pt ) const
{
	// First try for a fixed point.
	if ( !fixed_point( pt ) )
	{
		// No fixed point, so iterate to narrow in on the attractor.
		pt.pt().xy( 0, 0 );
		pt.colour( rgb( 0, 0, 0 ) );
		for ( int loop = 0; loop < 20; ++loop )
		{
			apply( pt, pt );
		}
	}
}


//
// Test if the rule has a fixed point.
//

bool
ifs_rule::has_fixed_point() const
{
	// See the fixed_point method for the derivation of this formula.
	return epsilon_nz( (1.0-m_a) * (1.0-m_d) - (m_b*m_c) );
}


//
// Find the fixed point of the rule.
//

bool
ifs_rule::fixed_point( ifs_point &pt ) const
{
	if ( m_rule_type != irt_linear )
	{
		return false;
	}

	// The following applies only to linear rules. Other types are not checked
	// for fixed points.
	//
	// To find the fixed point we start with the basic rule equations:
	//
	//   [1]  x' = ax + by + e
	//   [2]  y' = cx + dy + f
	//
	// Find the point where x' = x and y' = y. The equations above become:
	//
	//   [1]->[3]  x = ax + by + e
	//   [2]->[4]  y = cx + dy + f
	//
	// Rearranging these become:
	//
	//   [3]->[5]  (1-a).x = by + e
	//   [4]->[6]  (1-d).y = cx + f
	//
	// Eliminate y to find x:
	//
	//   [5]+[6]->[7]  (1-a).x = b.((cx + f) / (1-d)) + e
	//
	// Rearranging:
	//
	//   [7]->[8]  (1-a).(1-d).x = bcx + bf + e.(1-d)
	//   [8]->[9]  ((1-a).(1-d)-bc).x = bf + e.(1-d)
	//   [9]->[10]  x = (bf + e.(1-d)) / ((1-a).(1-d) - bc)
	//
	// Following the same process for y gives:
	//
	//   [11]  y = (ce + f.(1-a)) / ((1-a).(1-d) - bc)
	
	// First find the denominator.
	double denominator = (1.0-m_a) * (1.0-m_d) - (m_b*m_c);
	
	// If the denominator must be definitely non-zero for there to be an
	// attractor for this rule.
	if ( epsilon_nz( denominator ) )
	{
		// Calculate where the point is.
		pt.pt().x( (m_b*m_f + m_e*(1.0-m_d)) / denominator );
		pt.pt().y( (m_c*m_e + m_f*(1.0-m_a)) / denominator );
		pt.colour( m_colour );
		return true;
	}
	
	// No fixed point.
	return false;
}


//
// Test the contractivity of the rule.
//

ifs_rule_contractivity
ifs_rule::contractivity() const
{
	if ( m_rule_type != irt_linear )
	{
		return irc_unknown;
	}

	// The following applies only to linear rules. Other types are not checked
	// for contractivity.
	//
	// For a rule to be contractive the distance between any two points (other
	// than the fixed point) must decrease after they are passed through the
	// transformation:
	//
	//   dist( P, Q ) > dist( P', Q' )
	//
	// To simplify matters the constants in the rule equation can be ignored
	// since they do not change whether a rule contracts. This also has the
	// effect of moving the fixed point to (0, 0). To further simplify the
	// calculations let Q be (0, 0). Provided the rule has a fixed point Q' will
	// also be (0, 0). The equation above becomes:
	//
	//   [1]  x^2 + y^2 > (a^2 + b^2).x^2 + (c^2 + d^2).y^2 + 2.(ab + cd).x.y
	//
	// First consider cases on the axes. Let x <> 0 and y = 0:
	//
	//   [1]->[2]  x^2 > (a^2 + b^2).x^2
	//   [2]->[3]  a^2 + b^2 < 1
	//
	// And now let x = 0 and y <> 0:
	//
	//   [1]->[4]  y^2 > (c^2 + d^2).y^2
	//   [4]->[5]  c^2 + d^2 < 1
	//
	// Now consider the general case where x <> 0 and y = kx:
	//
	//   [1]->[6]  x^2 + k^2.x^2 > (a^2 + b^2).x^2 + (c^2 + d^2).k^2.x^2 + 2.(ab + cd).k.x^2
	//   [6]->[7]  (1 + k^2).x^2 > (c^2 + d^2).k^2.x^2 + 2.(ab + cd).k.x^2 + (a^2 + b^2).x^2
	//
	// Dividing both sides by x^2:
	//
	//   [7]->[8]  1 + k^2 > (c^2 + d^2).k^2 + 2.(ab + cd).k + (a^2 + b^2)
	//   [8]->[9]  1 > (c^2 + d^2 - 1).k^2 + 2.(ab + cd).k + (a^2 + b^2)
	//
	// This is the equation for a parabola. The only way it can remain less than
	// 1 is for the parabola to be inverted, meaning the coefficient on k^2 must
	// be negative, which in turn means c^2 + d^2 must be less than 1. This has
	// already been established as a requirement by [5]. The easiest way to
	// check whether the parabola remains below 1 is to calculate its value at
	// its apex, which is found by calculating the point with a 0 gradient, so
	// take the derivative:
	//
	//   [9]->[10]   2.(c^2 + d^2 - 1).k + 2.(ab + cd) = 0
	//   [10]->[11]  k = -(ab + cd) / (c^2 + d^2 - 1)

	// Make sure the rule has a fixed point.
	if ( !has_fixed_point() )
	{
		return irc_not_contractive;
	}

	// Check the axis conditions.
	double a2b2 = m_a*m_a + m_b*m_b;
	if ( !epsilon_pos( 1.0 - a2b2 ) )
	{
		return irc_not_contractive;
	}
	double c2d2 = m_c*m_c + m_d*m_d;
	if ( !epsilon_pos( 1.0 - c2d2 ) )
	{
		return irc_not_contractive;
	}
	
	// Find the apex of the parabola.
	double abcd = m_a*m_b + m_c*m_d;
	double k = -abcd/(c2d2 - 1.0);
	double v = (c2d2 - 1.0)*k*k + 2.0*abcd*k + a2b2;
	return epsilon_pos( 1.0 - v ) ? irc_contractive : irc_not_contractive;
}


//
// Fill a matrix with the rule coefficients.
//

void
ifs_rule::to_matrix( matrix3 &m3 ) const
{
	m3.set( 0, 0, m_a );
	m3.set( 0, 1, m_b );
	m3.set( 0, 2, m_e );
	m3.set( 1, 0, m_c );
	m3.set( 1, 1, m_d );
	m3.set( 1, 2, m_f );
	m3.set( 2, 0, 0.0 );
	m3.set( 2, 1, 0.0 );
	m3.set( 2, 2, 1.0 );
}


//
// Set the rule coefficients from a matrix.
//

void
ifs_rule::from_matrix( const matrix3 &m3 )
{
	m_a = m3.get( 0, 0 );
	m_b = m3.get( 0, 1 );
	m_e = m3.get( 0, 2 );
	m_c = m3.get( 1, 0 );
	m_d = m3.get( 1, 1 );
	m_f = m3.get( 1, 2 );
	assert( m3.get( 2, 0 ) == 0.0 );
	assert( m3.get( 2, 1 ) == 0.0 );
	assert( m3.get( 2, 2 ) == 1.0 );
}


}; // namespace eph
