// ephemera/ifs_rule.h
// A rule in an IFS fractal.
// Copyright (c) 2015 Ben Michell <http://www.muryan.com/>.

// This file is part of Ephemera.
//
// Ephemera is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Ephemera is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Ephemera.  If not, see <http://www.gnu.org/licenses/>.

#ifndef EPHEMERA_IFS_RULE_H
#define EPHEMERA_IFS_RULE_H


#include "colour.h"
#include "geometry.h"
#include "ifs_point.h"


namespace eph
{


// Forwards.
class matrix3;


//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


// Type of rules. The type determines which operations are applied to a point
// to transform it into a new point. The classic rule type is linear.
enum ifs_rule_type
{
	irt_linear,			// Linear.
	irt_trig,			// Sine and cosine.
	irt_trig_sq,		// Sine and cosine squared.
	irt_linear_trig,	// X is linear, Y is trig.
	irt_signed_sqrt,	// sign( n ) * sqrt( abs( n ) ).

	irt_mixture,		// Special value used by ifs_fractal::randomize.

	irt_count,						// Number of rule types.
	irt_normal_count = irt_mixture	// Number of non-mixture rule types.
};


//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


// Contractivity of a rule.
enum ifs_rule_contractivity
{
	irc_contractive,
	irc_not_contractive,
	irc_unknown
};


//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


// A rule in an IFS fractal.
class ifs_rule
{
public:

	ifs_rule( int weight, ifs_rule_type rule_type, const double &a, const double &b, const double &c, const double &d, const double &e, const double &f, const rgb &colour, const double &blend );
		// Construction. The weight has no special meaning for the rule itself,
		// and is intended for the user of this class. The rule_type specified
		// the type of transformation applied to the point. The values a through
		// f are the standard IFS coefficients. The colour and blend specify how
		// this rule affects the colour of the point. The output colour is:
		//   source_colour*(1-bend) + colour*blend

	void scale( const double &scale_x, const double &scale_y );
		// Scale the rule by the specified amounts. This only works with the
		// linear rule type.

	void rotate( const double &radians );
		// Rotate the rule by the specified angle. This only works with the
		// linear rule type.

	void shear( const double &x_in_y, const double &y_in_x );
		// Shear the rule by the specified amounts. This only works with the
		// linear rule type.

	void mirror_x() { scale( -1.0, 1.0 ); }
	void mirror_y() { scale( 1.0, -1.0 ); }
		// Mirror the rule in X or Y. This only works with the linear rule type.

	void apply( const ifs_point &pt, ifs_point &new_point ) const;
		// Apply the rule to the coloured point. It is safe to pass the same
		// point in as both the current and new points.

	void apply( const point &pt, point &new_pt ) const;
		// Apply the rule to the point. It is safe to pass the same point in as
		// both the current and new points.

	void start_point( ifs_point &pt ) const;
		// Find a starting point to use for this rule. If the rule has a fixed
		// point then this is used, otherwise a point on the attractor will be
		// determined empirically.

	bool has_fixed_point() const;
		// Test if the rule has a fixed point. This only works for linear rules.

	bool fixed_point( ifs_point &pt ) const;
		// Find the fixed point of the rule. This only works for linear rules.
		// The return value indicates whether a fixed point was found.

	ifs_rule_contractivity contractivity() const;
		// Test the contractivity of the rule. Only linear rules are tested.
		// Other types will return irc_unknown.

	int weight() const { return m_weight; }
		// Weighting for the rule.

	const double &a() { return m_a; }
	void a( const double &new_a ) { m_a = new_a; }
	const double &b() { return m_b; }
	void b( const double &new_b ) { m_b = new_b; }
	const double &c() { return m_c; }
	void c( const double &new_c ) { m_c = new_c; }
	const double &d() { return m_d; }
	void d( const double &new_d ) { m_d = new_d; }
	const double &e() { return m_e; }
	void e( const double &new_e ) { m_e = new_e; }
	const double &f() { return m_f; }
	void f( const double &new_f ) { m_f = new_f; }
		// Rule coefficients.

private:

	void to_matrix( matrix3 &m3 ) const;
		// Fill a matrix with the rule coefficients.

	void from_matrix( const matrix3 &m3 );
		// Set the rule coefficients from a matrix.

	int m_weight;
		// Weighting for the rule.

	ifs_rule_type m_rule_type;
		// Type of rule.

	double m_a;
	double m_b;
	double m_c;
	double m_d;
	double m_e;
	double m_f;
		// Rule coefficients. These are used to transform points as follows:
		//   x' = ax + by + e
		//   y' = cx + dy + f

	rgb m_colour;
		// Colour of the rule.

	unsigned char m_red[256];
	unsigned char m_green[256];
	unsigned char m_blue[256];
		// Colour modification tables. These tables are precomputed so that the
		// colour resulting from applying the rule is found simply by looking
		// up by the starting colour.
};


}; // namespace eph


#endif // EPHEMERA_IFS_RULE_H
