// ephemera/matrix3.cpp
// Simple matrix class used for transformations.
// Copyright (c) 2015 Ben Michell <http://www.muryan.com/>.

// This file is part of Ephemera.
//
// Ephemera is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Ephemera is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Ephemera.  If not, see <http://www.gnu.org/licenses/>.

#include "global.h"

#include "matrix3.h"


namespace eph
{


//
// Set the matrix to the identity matrix.
//

void
matrix3::identity()
{
	for ( size_t col = 0; col < 3; ++col )
	{
		for ( size_t row = 0; row < 3; ++row )
		{
			m_matrix[col][row] = (col == row) ? 1.0 : 0.0;
		}
	}
}


//
// Set the matrix to a scale transformation.
//

void
matrix3::scale( const double &sx, const double &sy )
{
	identity();
	m_matrix[0][0] = sx;
	m_matrix[1][1] = sy;
}


//
// Set the matrix to a rotate transformation.
//

void
matrix3::rotate( const double &radians )
{
	identity();
	m_matrix[0][0] = cos( radians );
	m_matrix[1][0] = sin( radians );
	m_matrix[0][1] = -m_matrix[1][0];
	m_matrix[1][1] = m_matrix[0][0];
}


//
// Set the matrix to a shear transformation.
//

void
matrix3::shear( const double &x_in_y, const double &y_in_x )
{
	identity();
	m_matrix[0][1] = x_in_y;
	m_matrix[1][0] = y_in_x;
}


//
// Multiply two matrices.
//

matrix3
matrix3::operator*( const matrix3 &m ) const
{
	matrix3	result;
	for ( size_t col = 0; col < 3; ++col )
	{
		for ( size_t row = 0; row < 3; ++row )
		{
			double	value = 0;
			for ( size_t accum = 0; accum < 3; ++accum )
			{
				value += m_matrix[accum][row] * m.m_matrix[col][accum];
			}
			result.m_matrix[col][row] = value;
		}
	}
	return result;
}


}; // namespace eph
