// ephemera/renderer.cpp
// Management of fractal drawing.
// Copyright (c) 2015 Ben Michell <http://www.muryan.com/>.

// This file is part of Ephemera.
//
// Ephemera is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Ephemera is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Ephemera.  If not, see <http://www.gnu.org/licenses/>.

#include "global.h"

#include "geometry.h"
#include "ifs_fractal.h"
#include "renderer.h"


namespace eph
{


// Number of points to provide in the point generation buffer.
const size_t renderer::m_point_buffer_size = 1000;


//
// Construction.
//

renderer::renderer( unsigned char *bitmap, int width, int height, int bytes_per_row ) :
	m_bitmap( bitmap ),
	m_width( width ), m_height( height ),
	m_bytes_per_row( bytes_per_row ),
	m_offset_x( 0.0 ), m_offset_y( 0.0 ), m_scale_x( 1.0 ), m_scale_y( 1.0 ),
	m_drawing_method( cs_overwrite ),
	m_point_buffer( m_point_buffer_size )
{
	assert( bitmap != 0 );
	assert( width > 0 );
	assert( height > 0 );
	assert( bytes_per_row >= (3*width) );
}


//
// Set the fractal to draw.
//

void
renderer::fractal( const ifs_fractal &ifs, const rectangle &rect, colour_system drawing_method )
{
	// Store the fractal and calculate the transformation between the
	// coordinates generated from the fractal and bitmap coordinates.
	m_fractal = ifs;
	m_drawing_method = drawing_method;
	m_offset_x = -rect.min_x();
	m_offset_y = -rect.min_y();
	m_scale_x = double( m_width ) / std::max( rect.width(), 1.0 );
	m_scale_y = double( m_height ) / std::max( rect.height(), 1.0 );
}


//
// Clear the bitmap to black.
//

void
renderer::clear()
{
	memset( m_bitmap, 0, m_height * m_bytes_per_row );
}


//
// Draw some points into the bitmap.
//

void
renderer::draw( long point_count )
{
	assert( !m_fractal.empty() );
	ifs_point pt;
	m_fractal.start_point( pt );
	for ( ; point_count > 0; point_count -= m_point_buffer.size() )
	{
		size_t points = std::min( size_t( point_count ), m_point_buffer.size() );
		m_fractal.generate_points( pt, m_point_buffer, points );
		pt = m_point_buffer[points-1];

		for ( std::vector<ifs_point>::const_iterator iter = m_point_buffer.begin(); iter != m_point_buffer.end(); ++iter )
		{
			int x = (int) floor( (iter->pt().x() + m_offset_x) * m_scale_x );
			int y = (int) floor( (iter->pt().y() + m_offset_y) * m_scale_y );
			if ( x >= 0 && x < m_width && y >= 0 && y < m_height )
			{
				unsigned char *pixel = m_bitmap + y*m_bytes_per_row + 3*x;

				switch ( m_drawing_method )
				{
				default:
					assert( !"Unhandled drawing method." );
					// Fall through.
				case cs_overwrite:
					pixel[0] = iter->colour().b();
					pixel[1] = iter->colour().g();
					pixel[2] = iter->colour().r();
					break;

				case cs_merge:
					pixel[0] = std::max( pixel[0], iter->colour().b() );
					pixel[1] = std::max( pixel[1], iter->colour().g() );
					pixel[2] = std::max( pixel[2], iter->colour().r() );
					break;

				case cs_intensity:
					if ( pixel[0] < iter->colour().b() )
					{
						pixel[0] += std::max( (iter->colour().b() - pixel[0]) >> 6, 1 );
					}
					if ( pixel[1] < iter->colour().g() )
					{
						pixel[1] += std::max( (iter->colour().g() - pixel[1]) >> 6, 1 );
					}
					if ( pixel[2] < iter->colour().r() )
					{
						pixel[2] += std::max( (iter->colour().r() - pixel[2]) >> 6, 1 );
					}
					break;
				}
			}
		}
	}
}


}; // namespace eph
