// ephemera/scheme.cpp
// Parameters for drawing a set of fractals.
// Copyright (c) 2015 Ben Michell <http://www.muryan.com/>.

// This file is part of Ephemera.
//
// Ephemera is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Ephemera is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Ephemera.  If not, see <http://www.gnu.org/licenses/>.

#include "global.h"

#include "ifs_fractal.h"
#include "random.h"
#include "scheme.h"


namespace eph
{


//
// Construction for standard scheme types.
//

scheme::scheme( unsigned int weight, unsigned int min_fractals, unsigned int max_fractals, bool contractive, colour_system drawing_method, double point_density, double min_scale, double max_scale, double centering ) :
	m_weight( weight ),
	m_type( st_standard ),
	m_min_fractals( min_fractals ), m_max_fractals( max_fractals ),
	m_clear_delay( 5 ),
	m_contractive( contractive ),
	m_dimension_filter( false ), m_min_dimension( 0.0 ),
	m_min_saturation( 0.8 ), m_min_brightness( 0.8 ),
	m_drawing_method( drawing_method ),
	m_point_density( point_density ),
	m_min_scale( min_scale ), m_max_scale( max_scale ),
	m_centering( centering )
{
}


//
// Construction for special scheme types.
//

scheme::scheme( unsigned int weight, scheme_type type, unsigned int clear_delay ) :
	m_weight( weight ),
	m_type( type ),
	m_clear_delay( clear_delay )
{
	assert( weight > 0 );
	assert( type == st_ferns || type == st_gasket );

	// Clear common values that aren't used.
	m_contractive = false;
	m_dimension_filter = false;
	m_min_dimension = 0.0;
	m_min_saturation = 0.0;
	m_min_brightness = 0.0;

	// Configuration is type-specific.
	if ( type == st_ferns )
	{
		m_min_fractals = 15;
		m_max_fractals = 25;
		m_drawing_method = cs_overwrite;
		m_point_density = 0.9;
		m_min_scale = 0.4;
		m_max_scale = 1.0;
		m_centering = 1.1;
	}
	else
	{
		assert( type == st_gasket );
		m_min_fractals = 8;
		m_max_fractals = 20;
		m_drawing_method = cs_merge;
		m_point_density = 0.5;
		m_min_scale = 0.3;
		m_max_scale = 1.0;
		m_centering = 0.9;
	}
}


//
// Check if the scheme is valid for use.
//

bool
scheme::valid() const
{
	if ( m_weight < 1 )
	{
		return false;
	}
	if ( m_min_fractals < 1 || m_min_fractals > m_max_fractals )
	{
		return false;
	}

	if ( m_type == st_standard && m_rule_ranges.empty() )
	{
		return false;
	}

	return true;
}


//
// Attempt to create a fractal from the scheme.
//

bool
scheme::create_fractal( const double &aspect_ratio, ifs_fractal &fractal, rectangle &rect, double &dimension, double &area ) const
{
	// Handle the special types.
	assert( epsilon_pos( aspect_ratio ) );
	if ( m_type == st_ferns )
	{
		create_fern_fractal( aspect_ratio, fractal, rect, dimension, area );
		return true;
	}
	else if ( m_type == st_gasket )
	{
		create_gasket_fractal( aspect_ratio, fractal, rect, dimension, area );
		return true;
	}

	// Handle the standard type.
	assert( !m_rule_ranges.empty() );
	size_t rr = (size_t) random::get( 0U, m_rule_ranges.size()-1 );
	fractal.randomize( m_rule_ranges[rr].type, m_rule_ranges[rr].min_rules, m_rule_ranges[rr].max_rules, m_contractive, m_min_saturation, m_min_brightness );
	fractal.analyze( rect, dimension );
	double raw_area = rect.width() * rect.height();
	rect.inflate_to_aspect_ratio( aspect_ratio );
	fractal.scale_and_move( rect, m_min_scale, m_max_scale, m_centering );
	double final_area = rect.width() * rect.height();
	area = epsilon_nz( final_area ) ? raw_area / final_area : 0.0;
	return !m_dimension_filter || dimension >= m_min_dimension;
}


//
// Create a fern fractal.
//

void
scheme::create_fern_fractal( const double &aspect_ratio, ifs_fractal &fractal, rectangle &rect, double &dimension, double &area ) const
{
	// Pick colours for the stalk and leaf parts of the fern.
	rgb stalk_colour;
	stalk_colour.r( (unsigned char) random::get( 0x00, 0x5a ) );
	stalk_colour.g( (unsigned char) random::get( 0x26, 0x5a ) );
	stalk_colour.b( 0 );
	rgb leaf_colour;
	leaf_colour.r( 0 );
	leaf_colour.g( (unsigned char) random::get( 0x4c, 0xff ) );
	leaf_colour.b( (unsigned char) random::get( 0x00, 0x72 ) );

	// Create the rules.
	fractal.clear();
	fractal.add_rule( ifs_rule( 6, irt_linear, 0.16, 0.0, 0.0, 0.0, 0.0, 0.0, stalk_colour, 1.00 ) );
	fractal.add_rule( ifs_rule( 10, irt_linear, 0.22, -0.23, 0.26, 0.2, -1.6, 0.0, leaf_colour, 0.15 ) );
	fractal.add_rule( ifs_rule( 10, irt_linear, 0.24, -0.26, -0.28, -0.15, -0.44, 0.0, leaf_colour, 0.15 ) );
	fractal.add_rule( ifs_rule( 74, irt_linear, 0.85, 0.04, -0.04, 0.75, -1.6, 0.0, leaf_colour, 0.05 ) );

	// Transform the fern and position it on the screen.
	fractal.rotate( random::get( -M_PI, M_PI ) );
	if ( random::get( 0, 1 ) == 1 )
	{
		fractal.mirror_x();
	}
	fractal.analyze( rect, dimension );
	double raw_area = rect.width() * rect.height();
	rect.inflate_to_aspect_ratio( aspect_ratio );
	fractal.scale_and_move( rect, m_min_scale, m_max_scale, m_centering );
	double final_area = rect.width() * rect.height();
	area = epsilon_nz( final_area ) ? raw_area / final_area : 0.0;
}


//
// Create a gasket fractal.
//

void
scheme::create_gasket_fractal( const double &aspect_ratio, ifs_fractal &fractal, rectangle &rect, double &dimension, double &area ) const
{
	// The method for constructing the 3, 5, and 6 gaskets is described in:
	//   The Pattern Book: Fractals, Art and Nature
	//   Clifford Pickover
	//   Pages 269-272
	//
	// n = number of sides
	// k = number of cut-outs
	// c = cut-out ratio (to shrunk side length)
	// f = shrink factor
	//
	// k = int( (n-1) / 4 )
	// c = sum i = 1 -> k, cos( 360 * i / n )
	// f = 1 / (2 * (1+c))
	//
	// The translations come from:
	//
	// i = 0 -> n-1
	//   x = cos( 90 + (360 / n) * i )
	//   y = sin( 90 + (360 / n) * i )

	// Templates for gasket rule creation (values for a, b, c, d, e, f).
	static const double	gasket_3[3][6] =
	{
		{ 0.5, 0, 0, 0.5, 0, 1 },
		{ 0.5, 0, 0, 0.5, -0.866, -0.5 },
		{ 0.5, 0, 0, 0.5, 0.866, -0.5 }
	};
	static const double	gasket_5[5][6] =
	{
		{ 0.382, 0, 0, 0.382, 0, 1 },
		{ 0.382, 0, 0, 0.382, -0.9511, 0.3090 },
		{ 0.382, 0, 0, 0.382, -0.5878, -0.8090 },
		{ 0.382, 0, 0, 0.382, 0.5878, -0.8090 },
		{ 0.382, 0, 0, 0.382, 0.9511, 0.3090 }
	};
	static const double	gasket_6[6][6] =
	{
		{ 0.3333, 0, 0, 0.3333, 0, 1 },
		{ 0.3333, 0, 0, 0.3333, -0.866, 0.5 },
		{ 0.3333, 0, 0, 0.3333, -0.866, -0.5 },
		{ 0.3333, 0, 0, 0.3333, 0, -1 },
		{ 0.3333, 0, 0, 0.3333, 0.866, -0.5 },
		{ 0.3333, 0, 0, 0.3333, 0.866, 0.5 }
	};

	// Create one of the gaskets.
	const double *rule_template = 0;
	size_t rule_count = 0;
	double which = random::get( 0.0, 1.0 );
	if ( which < 0.42 )
	{
		rule_template = gasket_3[0];
		rule_count = numberof( gasket_3 );
	}
	else if ( which < 0.75 )
	{
		rule_template = gasket_5[0];
		rule_count = numberof( gasket_5 );
	}
	else
	{
		rule_template = gasket_6[0];
		rule_count = numberof( gasket_6 );
	}

	// Create the rules.
	fractal.clear();
	for ( size_t r = 0; r < rule_count; ++r )
	{
		rgb colour;
		colour.from_hsv( random::get( 0.0, 1.0 ), random::get( 0.4, 1.0 ), random::get( 0.7, 1.0 ) );
		fractal.add_rule( ifs_rule( 1, irt_linear, rule_template[0], rule_template[1], rule_template[2], rule_template[3], rule_template[4], rule_template[5], colour, 1.00 ) );
		rule_template += 6;
	}

	// Transform the gasket and position it on the screen.
	fractal.shear( random::get( -0.707, 0.707 ), 0.0 );
	fractal.scale( random::get( 0.5, 1.0 ), random::get( 0.5, 1.0 ) );
	fractal.rotate( random::get( -M_PI, M_PI ) );
	fractal.analyze( rect, dimension );
	double raw_area = rect.width() * rect.height();
	rect.inflate_to_aspect_ratio( aspect_ratio );
	fractal.scale_and_move( rect, m_min_scale, m_max_scale, m_centering );
	double final_area = rect.width() * rect.height();
	area = epsilon_nz( final_area ) ? raw_area / final_area : 0.0;
}


//
// Set a range of rules for a rule type.
//

void
scheme::add_rule_range( ifs_rule_type type, unsigned int min_rules, unsigned int max_rules )
{
	assert( type >= 0 && type < irt_count );
	assert( min_rules > 0 && min_rules <= max_rules );

	rule_range rr;
	rr.type = type;
	rr.min_rules = min_rules;
	rr.max_rules = max_rules;
	m_rule_ranges.push_back( rr );
}


//
// Set fractal dimension filtering.
//

void
scheme::dimension_filter( double min_dimension )
{
	assert( min_dimension >= 0.0 && min_dimension <= 2.0 );
	m_dimension_filter = true;
	m_min_dimension = min_dimension;
}


//
// Minimum colour saturation (0-1).
//

void
scheme::min_saturation( double new_min_saturation )
{
	assert( new_min_saturation >= 0.0 && new_min_saturation <= 1.0 );
	m_min_saturation = new_min_saturation;
}


//
// Minimum colour brightness (0-1).
//

void
scheme::min_brightness( double new_min_brightness )
{
	assert( new_min_brightness >= 0.0 && new_min_brightness <= 1.0 );
	m_min_brightness = new_min_brightness;
}


}; // namespace eph
