// ephemera/scheme.h
// Parameters for drawing a set of fractals.
// Copyright (c) 2015 Ben Michell <http://www.muryan.com/>.

// This file is part of Ephemera.
//
// Ephemera is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Ephemera is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Ephemera.  If not, see <http://www.gnu.org/licenses/>.

#ifndef EPHEMERA_SCHEME_H
#define EPHEMERA_SCHEME_H


#include "geometry.h"
#include "ifs_rule.h"
#include "renderer.h"


namespace eph
{


// Forwards.
class ifs_fractal;


//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


// Scheme types.
enum scheme_type
{
	st_standard,	// Standard fully-random scheme.
	st_ferns,		// A variety of fern leaves.
	st_gasket		// A variety of 3, 5, and 6 gaskets.
};


//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


// Parameters for drawing a set of fractals.
class scheme
{
public:

	scheme( unsigned int weight, unsigned int min_fractals, unsigned int max_fractals, bool contractive, colour_system drawing_method, double point_density, double min_scale, double max_scale, double centering );
		// Construction for standard scheme types. This constructor sets most,
		// but not all, values. In particular the minimum and maximum rule
		// counts must be set for at least one rule type.

	scheme( unsigned int weight, scheme_type type, unsigned int clear_delay );
		// Construction for special scheme types.

	bool valid() const;
		// Check if the scheme is valid for use.

	unsigned int weight() const { return m_weight; }
		// Selection weighting for the scheme.

	bool create_fractal( const double &aspect_ratio, ifs_fractal &fractal, rectangle &rect, double &dimension, double &area ) const;
		// Attempt to create a fractal from the scheme. The area gives the
		// proportion of the target area covered by the fractal. The return
		// value indicates successful creation. The only reason this method
		// returns failure is if the created fractal failed the dimension
		// filter. In this case the caller can re-call this method as many times
		// as required, or many give up and use the returned fractal. The aspect
		// ratio of the target display area must be supplied so that the fractal
		// can be scaled and positioned appropriately.
		
	unsigned int min_fractals() const { return m_min_fractals; }
	unsigned int max_fractals() const { return m_max_fractals; }
		// Range of fractals to generate.

	unsigned int clear_delay() const { return m_clear_delay; }
	void clear_delay( unsigned int new_clear_delay ) { m_clear_delay = new_clear_delay; }
		// Number of seconds to wait after generating all fractals.

	void add_rule_range( ifs_rule_type type, unsigned int min_rules, unsigned int max_rules );
		// Set a range of rules for a rule type.

	void dimension_filter( double min_dimension );
		// Set fractal dimension filtering (0-2).

	void min_saturation( double new_min_saturation );
		// Minimum colour saturation (0-1).

	void min_brightness( double new_min_brightness );
		// Minimum colour brightness (0-1).

	colour_system drawing_method() const { return m_drawing_method; }
		// Method to use for drawing pixels.

	double point_density() const { return m_point_density; }
		// Number of points per pixel to generate.

private:

	// Configuration for a rule type.
	struct rule_range
	{
		ifs_rule_type type;
		unsigned int min_rules;
		unsigned int max_rules;
	};

	void create_fern_fractal( const double &aspect_ratio, ifs_fractal &fractal, rectangle &rect, double &dimension, double &area ) const;
		// Create a fern fractal.

	void create_gasket_fractal( const double &aspect_ratio, ifs_fractal &fractal, rectangle &rect, double &dimension, double &area ) const;
		// Create a gasket fractal.

	unsigned int m_weight;
		// Selection weighting for the scheme.

	scheme_type m_type;
		// Special scheme type.

	unsigned int m_min_fractals;
	unsigned int m_max_fractals;
		// Range of fractals to generate.

	unsigned int m_clear_delay;
		// Number of seconds to wait after generating all fractals.

	std::vector<rule_range> m_rule_ranges;
		// Number of rules per fractal for each type.

	bool m_contractive;
		// Whether to look for contractive rules.

	bool m_dimension_filter;
	double m_min_dimension;
		// Fractal dimension filtering (0-2). Note that as this value increase
		// it gets harder to find a matching fractal. At dimension 2 the only
		// valid fractal would be a solid rectangle.

	double m_min_saturation;
		// Minimum colour saturation (0-1).

	double m_min_brightness;
		// Minimum colour brightness (0-1).

	colour_system m_drawing_method;
		// Method to use for drawing pixels.

	double m_point_density;
		// Number of points per pixel to generate.

	double m_min_scale;
	double m_max_scale;
		// Range of scaling allowed.

	double m_centering;
		// Maximum centering offset. A value of 0 means none and 1 means the
		// edge of the screen.
};


}; // namespace eph


#endif // EPHEMERA_SCHEME_H
