// ephemera/screen_saver.cpp
// Generic Windows screen saver base class.
// Copyright (c) 2015 Ben Michell <http://www.muryan.com/>.

// This file is part of Ephemera.
//
// Ephemera is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Ephemera is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Ephemera.  If not, see <http://www.gnu.org/licenses/>.

#include "global.h"

#include "screen_saver.h"


namespace eph
{


// Name of the screen saver window classes.
const char *screen_saver::m_window_class_name = "muryan.com.screensaver";
const char *screen_saver::m_secondary_window_class_name = "muryan.com.screensaver2";


// ID to use for the animation timer.
UINT screen_saver::m_timer_id = 1;


// Milliseconds after starting to ignore keyboard and mouse input.
const DWORD screen_saver::m_dead_time = 600;


// Multiple monitor support.
HINSTANCE screen_saver::m_user_dll = 0;
screen_saver::eph_get_monitor_info_func screen_saver::m_get_monitor_info = 0;
screen_saver::eph_enum_display_monitors_func screen_saver::m_enum_display_monitors = 0;


//
// Construction.
//

screen_saver::screen_saver() :
	m_hwnd( 0 ),
	m_initialized( false ),
	m_frame_rate( 0 ),
	m_timer_running( false ),
	m_preview_mode( false ),
	m_start_time( 0 ),
	m_mouse_moved( false )
{
	// Attempt to load the monitor functions.
	if ( m_user_dll == 0 )
	{
		m_user_dll = LoadLibrary( "User32.dll" );
		if ( m_user_dll != 0 )
		{
#ifdef UNICODE
			m_get_monitor_info = (eph_get_monitor_info_func) GetProcAddress( m_user_dll, L"GetMonitorInfoW" );
#else // UNICODE
			m_get_monitor_info = (eph_get_monitor_info_func) GetProcAddress( m_user_dll, "GetMonitorInfoA" );
#endif // UNICODE
			m_enum_display_monitors = (eph_enum_display_monitors_func) GetProcAddress( m_user_dll, _T("EnumDisplayMonitors") );
		}
	}
}


//
// Run the screen saver app.
//

int
screen_saver::run_app( std::string cmd_line, int cmd_show, int frame_rate )
{
	// Break the command line apart.
	cmd_line.erase( 0, cmd_line.find_first_not_of( ' ' ) );
	size_t separator = cmd_line.find_first_of( " :" );
	std::string arg0 = cmd_line.substr( 0, separator );
	std::string arg1;
	if ( separator != cmd_line.npos )
	{
		arg1 = cmd_line.substr( separator+1 );
	}

	// Extract the mode from the command line.
	char mode = 0;
	if ( arg0.length() >= 1 )
	{
		mode = tolower( arg0[0] );
	}
	if ( (mode == '-' || mode == '/') && arg0.length() >= 2 )
	{
		mode = tolower( arg0[1] );
	}

	// Handle configure mode.
	int result = 0;
	if ( mode == 'c' || mode == 0 )
	{
		result = configure( GetForegroundWindow() );
	}
	// Handle normal screen saver mode.
	else if ( mode == 's' )
	{
		result = open_screen_saver( 0, cmd_show, frame_rate );
	}
	// Handle preview screen saver mode.
	else if ( mode == 'p' && !arg1.empty() && IsWindow( HWND( atoi( arg1.c_str() ) ) ) )
	{
		result = open_screen_saver( HWND( atoi( arg1.c_str() ) ), cmd_show, frame_rate );
	}
	// Anything else is an error.
	else
	{
		std::string err = "Invalid command line '" + cmd_line + "'.";
		MessageBox( HWND_DESKTOP, err.c_str(), "Screen Saver", MB_ICONERROR | MB_OK );
	}
	return result;
}


//
// Open the screen saver window.
//

int
screen_saver::open_screen_saver( HWND preview_parent, int cmd_show, int frame_rate )
{
	m_frame_rate = frame_rate;
	m_preview_mode = preview_parent != 0;

	// Register the window classes.
	HINSTANCE instance = GetModuleHandle( 0 );
	WNDCLASS wc;
	wc.style = 0;
	wc.lpfnWndProc = WNDPROC( window_proc );
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = instance;
	wc.hIcon = 0;
	wc.hCursor = 0;
	wc.hbrBackground = (HBRUSH) GetStockObject( BLACK_BRUSH );
	wc.lpszMenuName = 0;
	wc.lpszClassName = m_window_class_name;
	if ( RegisterClass( &wc ) == 0 )
	{
		return 0;
	}

	wc.style = 0;
	wc.lpfnWndProc = WNDPROC( secondary_window_proc );
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = instance;
	wc.hIcon = 0;
	wc.hCursor = 0;
	wc.hbrBackground = (HBRUSH) GetStockObject( BLACK_BRUSH );
	wc.lpszMenuName = 0;
	wc.lpszClassName = m_secondary_window_class_name;
	if ( RegisterClass( &wc ) == 0 )
	{
		return 0;
	}

	// Reset the keyboard input and mouse movement detector.
	m_start_time = GetTickCount();
	m_mouse_moved = false;

	// Create the window.
	DWORD style = preview_mode() ? WS_CHILD : WS_POPUP;
	int width, height;
	if ( preview_parent != 0 )
	{
		RECT rect;
		GetClientRect( preview_parent, &rect );
		width = rect.right;
		height = rect.bottom;
	}
	else
	{
		width = GetSystemMetrics( SM_CXSCREEN );
		height = GetSystemMetrics( SM_CYSCREEN );
	}
	m_hwnd = CreateWindowEx( WS_EX_TOPMOST, m_window_class_name, "", style, 0, 0, width, height, preview_parent, 0, instance, (LPVOID) this ); 
	if ( m_hwnd == 0 )
	{
		return 0;
	}
	ShowWindow( m_hwnd, cmd_show );
	if ( !preview_mode() )
	{
		SetCursor( 0 );
	}

	// Open dummy windows on secondary monitors if not in preview mode.
	if ( !preview_mode() )
	{
		std::vector<RECT> secondaries;
		find_secondary_monitors( secondaries );
		for ( std::vector<RECT>::const_iterator monitor = secondaries.begin(); monitor != secondaries.end(); ++monitor )
		{
			HWND secondary = CreateWindowEx( WS_EX_TOPMOST | WS_EX_NOACTIVATE, m_secondary_window_class_name, "", WS_POPUP, monitor->left, monitor->top, monitor->right - monitor->left, monitor->bottom - monitor->top, 0, 0, instance, (LPVOID) this ); 
			if ( secondary != 0 )
			{
				ShowWindow( secondary, SW_SHOWNOACTIVATE );
			}
		}
	}

	// Start a timer if animation is required.
	if ( frame_rate != 0 )
	{
		DWORD timer_period = std::max( 1000/frame_rate, USER_TIMER_MINIMUM );
		m_timer_running = SetTimer( m_hwnd, m_timer_id, timer_period, 0 ) != 0;
	}

	// Run the message pump until the window closes.
	MSG msg;
	while ( GetMessage( &msg, 0, 0, 0 ) )
	{
		TranslateMessage( &msg );
		DispatchMessage( &msg );
	}
	return msg.wParam;
}


//
// Stop the screen saver window.
//

void
screen_saver::close_screen_saver()
{
	if ( m_hwnd != 0 )
	{
		PostMessage( m_hwnd, WM_CLOSE, 0, 0 );
	}
}


//
// Centre a window over its parent.
//

void
screen_saver::centre_over_parent( HWND hwnd )
{
	RECT rect;
	GetWindowRect( hwnd, &rect );
	int child_width = rect.right - rect.left;
	int child_height = rect.bottom - rect.top;
	GetWindowRect( GetParent( hwnd ), &rect );
	int parent_width = rect.right - rect.left;
	int parent_height = rect.bottom - rect.top;
	int new_x = rect.left + (parent_width-child_width)/2;
	int new_y = rect.top + (parent_height-child_height)/2;
	SetWindowPos( hwnd, 0, new_x, new_y, 0, 0, SWP_NOACTIVATE|SWP_NOSIZE|SWP_NOZORDER );
}


//
// Window proc for the screen saver.
//

LONG APIENTRY
screen_saver::window_proc( HWND hwnd, UINT msg, UINT wparam, LONG lparam )
{
	screen_saver *saver = (screen_saver *) GetWindowLong( hwnd, GWL_USERDATA );
	switch ( msg )
	{
	case WM_CREATE:
		{
			CREATESTRUCT *cs = (CREATESTRUCT *) lparam;
			saver = (screen_saver *) cs->lpCreateParams;
			if ( !saver->handle_create( hwnd ) )
			{
				return -1;
			}
		}
		break;

	case WM_DESTROY:
		if ( saver != 0 )
		{
			saver->handle_destroy( hwnd );
		}
		PostQuitMessage( 0 );
		break;

	case WM_PAINT:
		if ( saver != 0 && saver->handle_paint( hwnd ) )
		{
			return 0;
		}
		break;

	case WM_TIMER:
		if ( saver != 0 )
		{
			saver->handle_timer( hwnd, wparam );
		}
		break;

	case WM_SETCURSOR:
		if ( saver != 0 && !saver->preview_mode() )
		{
			SetCursor( 0 );
			return TRUE;
		}
		break;

	case WM_SYSCOMMAND:
		// Prevent other screen savers appearing, and the window closing.
		if ( wparam == SC_SCREENSAVE || wparam == SC_CLOSE )
		{
			return 0;
		}
		break;

	case WM_MOUSEMOVE:
		if ( saver != 0 && saver->check_keyboard_and_mouse() && saver->mouse_moved_enough() )
		{
			saver->close_screen_saver();
		}
		break;

	case WM_LBUTTONDOWN:
	case WM_MBUTTONDOWN:
	case WM_RBUTTONDOWN:
	case WM_SYSKEYDOWN:
	case WM_KEYDOWN:
		if ( saver != 0 && saver->check_keyboard_and_mouse() )
		{
			saver->close_screen_saver();
		}
		break;

	default:
		break;
	}
	return DefWindowProc( hwnd, msg, wparam, lparam );
}


//
// Window proc for blanking secondary monitors.
//

LONG APIENTRY
screen_saver::secondary_window_proc( HWND hwnd, UINT msg, UINT wparam, LONG lparam )
{
	screen_saver *saver = (screen_saver *) GetWindowLong( hwnd, GWL_USERDATA );
	switch ( msg )
	{
	case WM_CREATE:
		SetWindowLong( hwnd, GWL_USERDATA, (LONG) ((CREATESTRUCT *) lparam)->lpCreateParams );
		break;

	case WM_DESTROY:
		PostQuitMessage( 0 );
		break;

	case WM_SETCURSOR:
		SetCursor( 0 );
		return TRUE;

	case WM_SYSCOMMAND:
		// Prevent other screen savers appearing, and the window closing.
		if ( wparam == SC_SCREENSAVE || wparam == SC_CLOSE )
		{
			return 0;
		}
		break;

	case WM_MOUSEMOVE:
		if ( saver != 0 && saver->check_keyboard_and_mouse() && saver->mouse_moved_enough() )
		{
			saver->close_screen_saver();
		}
		break;

	case WM_LBUTTONDOWN:
	case WM_MBUTTONDOWN:
	case WM_RBUTTONDOWN:
	case WM_SYSKEYDOWN:
	case WM_KEYDOWN:
		if ( saver != 0 && saver->check_keyboard_and_mouse() )
		{
			saver->close_screen_saver();
		}
		break;

	default:
		break;
	}
	return DefWindowProc( hwnd, msg, wparam, lparam );
}


//
// Handle WM_CREATE for the screen saver window.
//

bool
screen_saver::handle_create( HWND hwnd )
{
	SetWindowLong( hwnd, GWL_USERDATA, LONG( this ) );
	RECT rect;
	GetClientRect( hwnd, &rect );
	m_initialized = true;
	return saver_init( hwnd, rect.right, rect.bottom );
}


//
// Handle WM_DESTROY for the screen saver window.
//

void
screen_saver::handle_destroy( HWND hwnd )
{
	if ( m_timer_running )
	{
		KillTimer( hwnd, m_timer_id );
		m_timer_running = false;
	}

	m_hwnd = 0;
}


//
// Handle WM_PAINT for the screen saver window.
//

bool
screen_saver::handle_paint( HWND hwnd )
{
	if ( !m_initialized )
	{
		return false;
	}

	HDC hdc;
	RECT dirty;
	PAINTSTRUCT	ps;
	if ( GetUpdateRect( hwnd, &dirty, FALSE ) )
	{
		hdc = BeginPaint( hwnd, &ps );
		saver_draw( hdc, dirty );
		EndPaint( hwnd, &ps );
	}
	else
	{
		GetClientRect( hwnd, &dirty );
		hdc = GetDC( hwnd );
		saver_draw( hdc, dirty );
		ReleaseDC( hwnd, hdc );
	}

	return true;
}


//
// Handle WM_TIMER for the screen saver window.
//

void
screen_saver::handle_timer( HWND hwnd, int timer_id )
{
	if ( m_initialized && timer_id == m_timer_id )
	{
		if ( saver_animate() )
		{
			InvalidateRect( hwnd, 0, FALSE );
			UpdateWindow( hwnd );
		}
	}
}


//
// Test if the mouse has moved enough to stop the screen saver.
//

bool
screen_saver::mouse_moved_enough()
{
	if ( !m_mouse_moved )
	{
		GetCursorPos( &m_mouse_origin );
		m_mouse_moved = true;
		return false;
	}
	else
	{
		POINT pt;
		GetCursorPos( &pt );
		int x = abs( pt.x - m_mouse_origin.x );
		int y = abs( pt.y - m_mouse_origin.y );
		return (x+y) > 4;
	}
}


//
// Find the bounds of all secondary monitors.
//

void
screen_saver::find_secondary_monitors( std::vector<RECT> &monitors ) const
{
	monitors.clear();
	if ( m_enum_display_monitors != 0 && m_get_monitor_info != 0 )
	{
		// Get the monitor handles. We need to get handles so we can identify
		// the primary monitor and ignore it.
		std::vector<HMONITOR> raw_monitors;
		m_enum_display_monitors( 0, 0, &screen_saver::monitor_enum_proc, (LPARAM) &raw_monitors );
		
		// Convert the monitor handles into rects.
		for ( std::vector<HMONITOR>::const_iterator iter = raw_monitors.begin(); iter != raw_monitors.end(); ++iter )
		{
			MONITORINFO info;
			memset( &info, 0, sizeof( info ) );
			info.cbSize = sizeof( info );
			m_get_monitor_info( *iter, &info );
			if ( (info.dwFlags&MONITORINFOF_PRIMARY) == 0 )
			{
				monitors.push_back( info.rcMonitor );
			}
		}
	}
}


//
// Monitor enumeration proc.
//

BOOL CALLBACK 
screen_saver::monitor_enum_proc( HMONITOR monitor, HDC hdc, LPRECT rect, LPARAM user_data )
{
	assert( user_data != 0 );
	std::vector<HMONITOR> *monitors = reinterpret_cast<std::vector<HMONITOR> *>( user_data );

	MONITORINFO info;
	memset( &info, 0, sizeof( info ) );
	info.cbSize = sizeof( info );

	assert( m_get_monitor_info != 0 );
	m_get_monitor_info( monitor, &info );
	monitors->push_back( monitor );
	return TRUE;
}


}; // namespace eph
