// ephemera/screen_saver.h
// Generic Windows screen saver base class.
// Copyright (c) 2015 Ben Michell <http://www.muryan.com/>.

// This file is part of Ephemera.
//
// Ephemera is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Ephemera is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Ephemera.  If not, see <http://www.gnu.org/licenses/>.

#ifndef EPHEMERA_SCREEN_SAVER_H
#define EPHEMERA_SCREEN_SAVER_H


namespace eph
{


// Generic Windows screen saver base class.
class screen_saver : public uncopyable
{
public:

	screen_saver();
		// Construction.

	virtual ~screen_saver() {}
		// Destruction.

	int run_app( std::string cmd_line, int cmd_show, int frame_rate );
		// Run the screen saver app. This will interpret the command line and
		// run in the appropriate mode: configure, normal screen saver, or
		// preview. The frame_rate sets the desired number of frames per second
		// for any animation. A value of 0 means no animation.

	virtual int configure( HWND parent ) = 0;
		// Display the configuration dialog. The dialog should be run modally,
		// and this method should not return until the dialog has been closed.
		// This method must be provided by a derived class.

	virtual bool saver_init( HWND hwnd, int width, int height ) { return true; }
		// Initialize the screen saver. This will always be called before the
		// first call to the saver_draw or saver_animate methods. A return value
		// of true will start the screen saver and false will abort it.

	virtual void saver_draw( HDC hdc, const RECT &dirty ) = 0;
		// Draw the window while running as a screen saver. This method must be
		// provided by a derived class.

	virtual bool saver_animate() { return false; }
		// Perform animation processing. This method is only called if the
		// frame_rate passed to the run_app method was non-zero. If this method
		// wants to trigger a redraw then it should return true, otherwise
		// false.

	bool preview_mode() const { return m_preview_mode; }
		// Test whether the screen saver is running in preview mode. When in
		// preview mode the screen saver will ignore key presses and mouse
		// input. Normally these would cause the screen saver to stop running.

	int frame_rate() const { return m_frame_rate; }
		// Number of frames of animation per second.

	static void centre_over_parent( HWND hwnd );
		// Centre a window over its parent. This can be used to position the
		// configuration dialog nicely.

private:

	static LONG APIENTRY window_proc( HWND hwnd, UINT msg, UINT wparam, LONG lparam );
		// Window proc for the screen saver.

	static LONG APIENTRY secondary_window_proc( HWND hwnd, UINT msg, UINT wparam, LONG lparam );
		// Window proc for blanking secondary monitors.

	int open_screen_saver( HWND preview_parent, int cmd_show, int frame_rate );
		// Open the screen saver window. If the preview_parent is not 0 then the
		// screen saver should run in preview mode, otherwise it should run
		// normally.

	void close_screen_saver();
		// Stop the screen saver window.

	bool handle_create( HWND hwnd );
		// Handle WM_CREATE for the screen saver window. Returns true if
		// successful.

	void handle_destroy( HWND hwnd );
		// Handle WM_DESTROY for the screen saver window.

	bool handle_paint( HWND hwnd );
		// Handle WM_PAINT for the screen saver window.

	void handle_timer( HWND hwnd, int timer_id );
		// Handle WM_TIMER for the screen saver window.

	bool check_keyboard_and_mouse() const { return !preview_mode() && (GetTickCount() - m_start_time) > m_dead_time; }
		// Test whether keyboard and mouse input cause the screen saver to stop.
		// While the screen saver is starting key presses and mouse input is
		// ignored for a short time.

	bool mouse_moved_enough();
		// Test if the mouse has moved enough to stop the screen saver.

	void find_secondary_monitors( std::vector<RECT> &monitors ) const;
		// Find the bounds of all secondary monitors. The monitors are returned
		// in no particular order.

	static BOOL CALLBACK monitor_enum_proc( HMONITOR monitor, HDC hdc, LPRECT rect, LPARAM user_data );
		// Monitor enumeration proc.

	static const char *m_window_class_name;
	static const char *m_secondary_window_class_name;
		// Name of the screen saver window classes.

	static UINT m_timer_id;
		// ID to use for the animation timer.

	static const DWORD m_dead_time;
		// Milliseconds after starting to ignore keyboard and mouse input. This
		// prevents the screen saver from being immediately terminated by the
		// action that triggered it, such as pressing the Control Panel Preview
		// button to set the screen saver running.

	HWND m_hwnd;
		// Window handle of the screen saver. If this is 0 if the screen saver
		// is not running.

	bool m_initialized;
		// Flag when the screen saver has been told to initialize. Before this
		// no calls to saver_draw or saver_animate are allowed.

	int m_frame_rate;
		// Number of frames of animation per second.

	bool m_timer_running;
		// Whether the animation timer is running.

	bool m_preview_mode;
		// Whether the screen saver is running in preview mode.

	DWORD m_start_time;
		// Time when the screen saver was started.

	bool m_mouse_moved;
	POINT m_mouse_origin;
		// Track the mouse for unblanking.

	typedef BOOL (WINAPI *eph_get_monitor_info_func)( HMONITOR, LPMONITORINFO );
	typedef BOOL (WINAPI *eph_enum_display_monitors_func)( HDC, LPCRECT, MONITORENUMPROC, LPARAM );
	static HINSTANCE m_user_dll;
	static eph_get_monitor_info_func m_get_monitor_info;
	static eph_enum_display_monitors_func m_enum_display_monitors;
		// Multiple monitor support.
};


}; // namespace eph


#endif // EPHEMERA_SCREEN_SAVER_H
